#region Using directives

using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Reflection;

using Framework.DxO;
using Framework.DxO.Annotation;
using Framework.DxO.Converter;

#endregion

namespace Framework.DxO.Impl
{
    /// <summary>
    /// IDataExchange{S,D}C^tF[X̕WNX񋟂܂
    ///
    /// ݂̌^ɓÕvpeB݂ĂꍇAȉ̌^̃vpeB͎Iɕϊ܂
    ///
    /// 1.v~eBu^
    /// 3.^
    /// 4.z
    /// 5.IListAIDisctionary
    /// 6.ICollection{T}AIList{T}AIDisctionary{K, V}
    /// 7.̑̋ۃNX
    /// 8.̑A^Ro[^ɂA݂Ɍ^ϊ\Ȍ^
    ///
    /// </summary>
    /// <author>Kazz.satou</author>
    /// <version> 1.1 11/30, 2006</version>
    ///
    public class DataExchangeImpl<S, D> : IDataExchange<S, D>
        where S : class
        where D : class
    {
        #region fields

        protected IDictionary<string, DxOMappingAttribute> dxoMapping;

        #endregion
        #region IDataExchange{S,D} member

        #region events

        public event ConvertEventHandler PrepareConvert;
        public event ConvertEventHandler ConvertCompleted;
        public event ConvertEventHandler ConvertFail;

        #endregion

        public virtual void Convert(S source, D dest)
        {
            this.AssignTo(source, dest);
        }

        public virtual D Convert(S source)
        {
            if (typeof(D).IsInterface || typeof(D).IsAbstract )
            {
                throw new DxOException(
                    string.Format(
                    @"̃\bh̓CX^X𐶐̂ŁA
                ߂ľ^ɃC^tF[X⒊ۃNX͎gȂ ^ : {0}"
                                  , typeof(D).Name));
            }
            D dest = Activator.CreateInstance(typeof (D)) as D;
            this.AssignTo(source, dest);
            return dest;
        }

        #endregion
        #region protected member

        /// <summary>
        /// ^T1^T2̃IuWFNgւ̃ATC{܂
        /// </summary>
        /// <param name="source">ATC̃IuWFNg</param>
        /// <param name="dest">ATC̃IuWFNg</param>
        protected virtual void AssignTo(S source, D dest)
        {
            //Ame[Vޔ
            this.CollectDxoMappingsAnnotation();

#if DxODEBUG
                Debug.WriteLine("  ");
                Debug.WriteLine("###   DxO Type Property Conversion <" + typeof(S).Name + "> To <" + typeof(D).Name + "> ###");
#endif
                Type destType = typeof(D);
                foreach (PropertyInfo info in typeof(S).GetProperties())
                {
                    this.TryExchangeSameNameProperty(info, source, dest, destType);
                }
#if DxODEBUG
                Debug.WriteLine("  ");
                Debug.WriteLine("###   DxO Type Property Conversion <" + typeof(S).Name + "> To <" + typeof(D).Name + "> terminated ###");
#endif
        }
        #endregion
        #region private member

        /// <summary>
        /// Cӂ̃vpeBΏۂ̃IuWFNg̃vpeBɕϊ܂
        /// </summary>
        /// <param name="sourceInfo">ΏۂƂȂĂvpeBZbg</param>
        /// <param name="sourceObj">ΏۂƂȂĂvpeBIuWFNgZbg</param>
        /// <param name="destObj">ϊΏۂ̃IuWFNgZbg</param>
        /// <param name="destType">ϊΏۂ̃IuWFNǧ^Zbg</param>
        private bool TryExchangeSameNameProperty(PropertyInfo sourceInfo, object sourceObj
                                      , object destObj, Type destType)
        {
            Debug.Assert(sourceInfo != null, " sourceInfo͔null̂͂");
            Debug.Assert(sourceObj != null, " sourceObj͔null̂͂");
            Debug.Assert(destObj != null, " destObj͔null̂͂");

            try
            {
                PropertyInfo destInfo;
                string targetPropertyName = sourceInfo.Name;
                bool matchDxoMapping = this.dxoMapping.ContainsKey(sourceInfo.Name);

                if ( matchDxoMapping )
                {
                    string targetName = this.dxoMapping[sourceInfo.Name].TargetPropertyName;
                    targetPropertyName = String.IsNullOrEmpty(targetName)
                                             ? sourceInfo.Name
                                             : targetName;
                }
                //ÕvpeBϊ擯vpeBɂ邩?
                destInfo = destType.GetProperty(targetPropertyName);
                if (destInfo != null && destInfo.CanRead && destInfo.CanWrite )
                {
                    object sourceValue = sourceInfo.GetValue(sourceObj, null);
                    object destValue = destInfo.GetValue(destObj, null);

#if DxODEBUG
                    Debug.WriteLine("  ");
                    Debug.WriteLine(string.Format(@"DxO exchange property source  = ""{0}"" Type : {1}"
                                                  , sourceInfo.Name, sourceInfo.PropertyType.Name));
                    Debug.WriteLine(string.Format("                        value = {0}"
                                                  , (( sourceValue != null) ? sourceValue:"null")));
#endif
                    if (sourceValue != null)
                    {
                        IPropertyConverter propertyConverter;
                        if ( matchDxoMapping )
                        {
                            DxOMappingAttribute attr = this.dxoMapping[sourceInfo.Name];
                            //Ro[^Ame[V̓t@Ng擾
                            propertyConverter = (attr.Ignore)
                                        ? null
                                        : (attr.PropertyConverter != null)
                                           ? Activator.CreateInstance(attr.PropertyConverter) as IPropertyConverter
                                           : ConverterFactory.GetConverter(sourceValue, destInfo.PropertyType);
#if DxODEBUG
                            if (attr.Ignore)
                            {
                                Debug.WriteLine("  ");
                                Debug.WriteLine(string.Format(@"DxO exchange property destination  = ""{0}"" Type : {1}"
                                                              , destInfo.Name, destInfo.PropertyType.Name));
                                Debug.WriteLine(@"                        DxOMapping annotation set """"Ignore = true""");
                            }
#endif
                        }
                        else
                        {
                            //Ro[^t@Ng擾
                            propertyConverter =  ConverterFactory.GetConverter(sourceValue, destInfo.PropertyType);
                        }

                        if (propertyConverter != null)
                        {
                            //Ro[^̃CxgA^b`s
                            this.AttachConverterEvent(propertyConverter);

                            try
                            {
                                //Ro[^ɂϊ
                                propertyConverter.Convert(sourceInfo.Name, sourceValue
                                                          , ref destValue, destInfo.PropertyType);
                                destInfo.SetValue(destObj, destValue, null);
#if DxODEBUG
                                Debug.WriteLine("                     |");
                                Debug.WriteLine("\t++ " + propertyConverter.GetType().Name + " use conversion ++");
                                Debug.WriteLine("                     |");
                                Debug.WriteLine(string.Format(@"                      destiny = ""{0}"" Type : {1}"
                                                              , destInfo.Name, destInfo.PropertyType.Name));
                                Debug.WriteLine(string.Format("                        value = {0}"
                                                              , ((destValue != null) ? destValue : "null")));
#endif
                                return true;
                            }
                            finally
                            {
                                //Ro[^̃Cxgf^b`s
                                this.DetachConverterEvent(propertyConverter);
                            }
                        }
                    }
                }
#if DxODEBUG
                if (destInfo != null)
                {
                     if ( !destInfo.CanRead )
                     {
                         Debug.WriteLine("  ");
                         Debug.WriteLine(string.Format(@"DxO exchange property destination  = ""{0}"" Type : {1}"
                                                       , destInfo.Name, destInfo.PropertyType.Name));
                         Debug.WriteLine("                        can't read value (no getter)");
                     }
                     if ( !destInfo.CanWrite)
                     {
                         Debug.WriteLine("  ");
                         Debug.WriteLine(string.Format(@"DxO exchange property destination  = ""{0}"" Type : {1}"
                                                       , destInfo.Name, destInfo.PropertyType.Name));
                         Debug.WriteLine("                        can't write value (no setter)");
                     }
                }
#endif
                return false;
            }
            catch (Exception e)
            {
                throw new DxOException(e.Message, e);
            }
        }
	 
        /// <summary>
        /// Ro[^̃CxgA^b`܂
        /// </summary>
        /// <param name="converter">݂̃Ro[^Zbg</param>
        private void AttachConverterEvent(IPropertyConverter converter)
        {
            converter.PrepareConvert += this.PrepareConvert;
            converter.ConvertCompleted += this.ConvertCompleted;
            converter.ConvertFail += this.ConvertFail;
        }
        /// <summary>
        /// Ro[^̃Cxgf^b`܂
        /// </summary>
        /// <param name="converter">݂̃Ro[^Zbg</param>
        private void DetachConverterEvent(IPropertyConverter converter)
        {
            converter.PrepareConvert -= this.PrepareConvert;
            converter.ConvertCompleted -= this.ConvertCompleted;
            converter.ConvertFail -= this.ConvertFail;
        }
        /// <summary>
        /// DxoMappingp̃Ame[VSĎ擾āAɕێ܂
        /// </summary>
        private void CollectDxoMappingsAnnotation()
        {
            if ( this.dxoMapping == null )
            {
                this.dxoMapping = new Dictionary<string, DxOMappingAttribute>();
                lock(this.dxoMapping)
                {
                    Attribute[] attrs = Attribute.GetCustomAttributes(this.GetType()
                                                       , (typeof(DxOMappingAttribute)));
                    foreach (Attribute attr in attrs)
                    {
                        DxOMappingAttribute dxoMapping = attr as DxOMappingAttribute;
                        if (!String.IsNullOrEmpty(dxoMapping.PropertyName))
                        {
                            this.dxoMapping.Add(dxoMapping.PropertyName, dxoMapping);
                        }
                    }

                }
            }

        }
        #endregion
    }
}